# -* coding: utf-8 *-
"""
:py:mod:`pynspector.module_inspections`
---------------------------------------
Main module for inspecting modules
"""
# System imports
from inspect import isfunction, getmembers

# Third-party imports
# Local imports


__all__ = ['get_module_functions']


def get_module_functions(module):
    """Get functions for a given module

    Return all public functions that are available on a module.

    Example:
    >>> from pynspector import module_inspections_fixture_test
    >>> functions = get_module_functions(module_inspections_fixture_test)
    >>> list(functions)
    >>> [module_inspections_fixture_test.dummy_func]
    >>> type(functions)
    >>> <type 'generator'>

    :param module module: Module to inspect for functions
    :return: Generator that returns functions that are available under this module
    :rtype: generator
    """

    return (
        func for name, func in getmembers(module)
        if not name.startswith("_") and isfunction(func)
    )



# -* coding: utf-8 *-
"""
:py:mod:`pynspector.models`
---------------------------
Here you will find all models that are used across pynspector.

What is a model?
================
A model is just a class that holds some attributes and is returned as a result of an operation.
"""
# System imports
import inspect

# Third-party imports
# Local imports


class Argument(object):
    """Argument object

    This object represents an argument that is passed to a function.
    Whenever you inspect a function, you will get a function object with all it's arguments in this format.
    """

    def __init__(self, name, default, kind, description, is_arg, position):
        """ Initialize Argument object

        :param str name: Name of the argument
        :param default: Default value for this argument
        :param str kind: The type of this argument
        :param str description: Description for this argument
        :param bool is_arg: Returns if the argument is arg or kwarg (keyword argument)
        :param int position: Position in arg list
        """
        self.name = name
        self.default = default
        self.kind = kind
        self.description = description
        self.is_arg = is_arg
        self.position = position

    @property
    def is_kwarg(self):
        return not self.is_arg

    @property
    def is_mandatory(self):
        """Returns if the argument is mandatory or not"""
        return self.is_arg


class Function(object):
    """Function object

    This object represents a function
    """

    def __init__(self, name, short_description, long_description, func, arguments):
        self.name = name
        self.short_description = short_description
        self.long_description = long_description
        self.func = func
        self.source_code = inspect.getsource(func)
        self.arguments = arguments or []



# -* coding: utf-8 *-
"""
:py:mod:`pynspector.func_inspections`
-------------------------------------
Main module for inspecting functions
"""
# System imports
import inspect
import six

# Third-party imports
# Local imports
from . import models
from .doc_parsers import sphinx_doc_parser


__all__ = ['get_function_args', 'get_default_args']


def get_default_args(func):
    """ Get default arguments for a function

    Return default arguments for a given function, if function has no default arguments, it
    returns an empty dictionary.

    Example:
    >>> def func(foo, bar=None):
    ...     pass
    >>> get_default_args(func)
    >>> {'bar': None}

    :param function func: Function to get default arguments
    :return: Dictionary with argument as key and default value as value
    :rtype: dict
    """
    if six.PY3:
        func_sign = inspect.signature(func)
        return {param.name: param.default
                for param in func_sign.parameters.values()
                if param.default is not inspect._empty}
    args, _, _, defaults = inspect.getargspec(func)
    if not defaults:
        return {}
    return dict(zip(args[-len(defaults):], defaults))


def get_function_args(func):
    """ Get all the arguments defined on a function

    Example:
    >>> def example_func(foo, bar=None):
    >>>    return foo, bar
    >>> get_function_args(example_func)
    >>> ['foo', 'bar']

    :param func func: Function to inspect
    :return: List of arguments
    :rtype: list
    """
    if six.PY3:
        func_sign = inspect.signature(func)
        return list(func_sign.parameters.keys())
    arg_spec = inspect.getargspec(func)
    return arg_spec.args


def get_func_inspect_result(func, doc_parser=sphinx_doc_parser):
    """Get inspect results for a function

    :param function func: Function to inspect
    :return: Object with all the information related with the function
    :rtype: models.Function
    """
    arguments = []
    name = func.__name__
    args = get_function_args(func)
    default_args = get_default_args(func)
    short_description, long_description, doc_args, returns = doc_parser(func.__doc__)
    for position, arg in enumerate(args):
        doc_arg = doc_args.get(arg) or {}
        kind = doc_arg.get('type')
        description = doc_arg.get('doc')
        default = default_args.get(arg)
        is_arg = arg not in default_args  # if it doesn't have a default value, then it's an argument
        # name, default_value, kind, description, is_arg, position
        arguments.append(
            models.Argument(name=arg, default=default, kind=kind,
                            description=description, is_arg=is_arg, position=position)
        )
    return models.Function(name=name, short_description=short_description,
                           long_description=long_description, func=func,
                           arguments=arguments)



# -* coding: utf-8 *-
"""
:py:mod:`pynspector.doc_parser`
-------------------------------
Here you will find parsers for docstrings, this will help you to retrieve all
arguments, types and descriptions defined on docstrings.

You can use different parsers such as:
- sphinx_doc_parser
"""
# System imports
import re
import sys

# Third-party imports
# Local imports


__all__ = ['sphinx_doc_parser']


PARAM_OR_RETURNS_REGEX = re.compile(r":(?:param|returns|return)")
RETURNS_REGEX = re.compile(
    r":(returns|return): (?P<doc>.*)(?:(?=:param)|(?=:raises)|(?=:rtype)|(?=:type)|\Z)", re.S
)
TYPE_REGEX = re.compile(
    r":type (?P<name>[\*\w]+): (?P<type>.*?)(?:(?=:param)|"
    r"(?=:return)|(?=:returns)|(?=:raises)|(?=:rtype)|\Z)", re.S
)
PARAM_REGEX = re.compile(
    r":param (?P<type>.*?)(?P<name>[\*\w]+): (?P<doc>.*?)(?:(?=:param)|"
    r"(?=:return)|(?=:returns)|(?=:raises)|(?=:rtype)|(?=:type)|\Z)", re.S
)


def _trim(docstring):
    """trim function from PEP-257

    :param str docstring: Docstring in string format
    :return: Docstring parsed with trim.
    :rtype: str
    """
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    indent = sys.maxsize
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < sys.maxsize:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)

    # Current code/unittests expects a line return at
    # end of multiline docstrings
    # workaround expected behavior from unittests
    if "\n" in docstring:
        trimmed.append("")

    # Return a single string:
    return "\n".join(trimmed)


def _reindent(string):
    """Reindent string"""
    return "\n".join(l.strip() for l in string.strip().split("\n"))


def sphinx_doc_parser(docstring):
    """Parse docstring and return short, long and arguments for this function
    
    This parser should work as sphinx does by default. It should be able to parse
    docstrings following next format:

    ```
    Title of the docstring

    Long description goes here, with
    multiline support.

    :param str argument_one: Argument one description
    :param argument_two: Argument two description
    :type argument_two: str
    :return: What the function returns
    ```

    - The short description or title is mandatory (at least for now).
    - The long description is optional, it will report "" if there's no long description available.
    - The param type is not mandatory, if there's no type for a param, it will report it as None.
    - The return statement could be also "returns", both works fine.

    :param str docstring: Docstring in string format
    :returns: Tuple with short_description, long_description, params, returns
    :rtype: tuple
    """
    short_description = long_description = returns = ""
    params = {}

    if docstring:
        docstring = _trim(docstring)

        lines = docstring.split("\n", 1)
        short_description = lines[0]

        if len(lines) > 1:
            long_description = lines[1].strip()

            params_returns_desc = None

            match = PARAM_OR_RETURNS_REGEX.search(long_description)
            if match:
                long_desc_end = match.start()
                params_returns_desc = long_description[long_desc_end:].strip()
                long_description = long_description[:long_desc_end].rstrip()

            if params_returns_desc:
                params = {}
                for type_, name, doc in PARAM_REGEX.findall(params_returns_desc):
                    params[name] = {'doc': ' '.join(_trim(doc).split('\n')),
                                    'type': type_.strip() if type_ else None}

                match = RETURNS_REGEX.search(params_returns_desc)
                if match:
                    returns = _reindent(match.group("doc"))

                for name, type_ in TYPE_REGEX.findall(params_returns_desc):
                    type_ = type_.strip()
                    if name in params:
                        params[name].update({'type': type_})
                    else:
                        params[name] = {'doc':'', 'type': type_}

    return short_description, long_description, params, returns
